@@ -1,4 +1,5 @@
-/*
+/* -*-mode:c;coding:utf-8; c-basic-offset:2;fill-column:70;c-file-style:"gnu"-*-
+ *
  * Copyright (C) 2009 Arnaud "arnau" Fontaine <arnau@mini-dweeb.org>
  *
  * This  program is  free  software: you  can  redistribute it  and/or
@@ -26,6 +27,8 @@
 #include <stdbool.h>
 
 #include <xcb/xcb.h>
+#include <xcb/xfixes.h>
+#include <xcb/randr.h>
 
 void display_init_event_handlers(void);
 
@@ -45,4 +48,6 @@ void display_init_redirect_finalise(void);
 void display_add_damaged_region(xcb_xfixes_region_t *, bool);
 void display_reset_damaged(void);
 
+void display_set_screen_refresh_rate(xcb_randr_get_screen_info_cookie_t);
+
 #endif
