@@ -1,4 +1,5 @@
-/*
+/* -*-mode:c;coding:utf-8; c-basic-offset:2;fill-column:70;c-file-style:"gnu"-*-
+ *
  * Copyright (C) 2009 Arnaud "arnau" Fontaine <arnau@mini-dweeb.org>
  *
  * This  program is  free  software: you  can  redistribute it  and/or
@@ -1051,6 +1052,7 @@ expose_destructor(void)
 plugin_vtable_t plugin_vtable = {
   .name = "expose",
   .events = {
+    NULL,
     NULL,
     NULL,
     expose_event_handle_key_release,
