@@ -1,4 +1,5 @@
-/*
+/* -*-mode:c;coding:utf-8; c-basic-offset:2;fill-column:70;c-file-style:"gnu"-*-
+ *
  * Copyright (C) 2009 Arnaud "arnau" Fontaine <arnau@mini-dweeb.org>
  *
  * This  program is  free  software: you  can  redistribute it  and/or
@@ -225,6 +226,10 @@ event_handle_startup(xcb_generic_event_t *event)
     }
 }
 
+/** Maximum number  of DamageNotify events received  before repainting
+    the full window */
+#define DAMAGE_NOTIFY_MAX 24
+
 /** Handler for DamageNotify events
  *
  * \param event The X DamageNotify event
@@ -241,51 +246,60 @@ event_handle_damage_notify(xcb_damage_notify_event_t *event)
 	(uintmax_t) event->geometry.width, (uintmax_t) event->geometry.height,
 	(uintmax_t) event->geometry.x, (uintmax_t) event->geometry.y);
 
+#ifdef __DEBUG__
+  static unsigned int damage_notify_event_counter = 0;
+  debug("DamageNotify: COUNT: %u", ++damage_notify_event_counter);
+#endif
+
   window_t *window = window_list_get(event->drawable);
-
-  /* The window may have disappeared in the meantime */
-  if(!window)
-    {
-      debug("Window %jx has disappeared", (uintmax_t) event->drawable);
-      return;
-    }
-
-  if(!window_is_visible(window))
-    {
-      xcb_damage_subtract(globalconf.connection, window->damage, XCB_NONE, XCB_NONE);
-      debug("Ignore damaged as Window %x is not visible", window->id);
-      return;
-    }
-
   xcb_xfixes_region_t damaged_region;
   bool is_temporary_region = false;
 
+  /* The window may have disappeared in the meantime or is not visible
+     so do nothing */
+  if(!window || !window_is_visible(window))
+    return;
   /* If the Window has never been  damaged, then it means it has never
      be painted on the screen yet, thus paint its entire content */
-  if(!window->damaged)
+  else if(!window->damaged)
     {
       damaged_region = window->region;
       window->damaged = true;
+      window->damaged_ratio = 1.0;
+    }
+  /* Do nothing if the window is already fully damaged */
+  else if(window->damaged_ratio >= WINDOW_FULLY_DAMAGED_RATIO)
+    {
+      debug("Window %jx fully damaged (cached)", (uintmax_t) window->id);
+      return;
+    }
+  /* If  the   window  is  considered   fully  damaged  or   too  many
+     DamageNotify  events   have  been   received,  then   repaint  it
+     completely */
+  else if(window->damage_notify_counter++ > DAMAGE_NOTIFY_MAX ||
+          window_get_damaged_ratio(window, event) >= WINDOW_FULLY_DAMAGED_RATIO)
+    {
+      debug("Window %jx damaged ratio: %.2f, counter: %d",
+           (uintmax_t) window->id,
+           window->damaged_ratio,
+           window->damage_notify_counter);
 
-      xcb_damage_subtract(globalconf.connection, window->damage, XCB_NONE, XCB_NONE);
+      /* @todo:  Perhaps  xcb_damage_add()  could  be  used  to  avoid
+         further events to  be sent as the window  is considered fully
+         damaged? */
+      damaged_region = window->region;
+      window->damaged_ratio = 1.0;
     }
   /* Otherwise, just paint the damaged Region (which may be the entire
      Window or part of it */
   else
     {
       damaged_region = xcb_generate_id(globalconf.connection);
-      
+
+      event->area.x += event->geometry.x;
+      event->area.y += event->geometry.y;
       xcb_xfixes_create_region(globalconf.connection, damaged_region,
-                               0, NULL);
-
-      xcb_damage_subtract(globalconf.connection, window->damage, XCB_NONE,
-                          damaged_region);
-
-      xcb_xfixes_translate_region(globalconf.connection, damaged_region,
-                                  (int16_t) (window->geometry->x +
-                                             window->geometry->border_width),
-                                  (int16_t) (window->geometry->y +
-                                             window->geometry->border_width));
+                               1, &event->area);
 
       is_temporary_region = true;
     }
@@ -295,6 +309,25 @@ event_handle_damage_notify(xcb_damage_notify_event_t *event)
   PLUGINS_EVENT_HANDLE(event, damage, window);
 }
 
+/** Handler for RRScreenChangeNotify events reported when the screen
+ *  configuration change and is meaningful to get the new refresh rate
+ *
+ * \param event The X RRScreenChangeNotify event
+ */
+static void
+event_handle_randr_screen_change_notify(xcb_randr_screen_change_notify_event_t *event)
+{
+  debug("RandrScreenChangeNotify: root=%jx", (uintmax_t) event->root);
+
+  xcb_randr_get_screen_info_cookie_t cookie =
+    xcb_randr_get_screen_info(globalconf.connection,
+                              globalconf.screen->root);
+
+  display_set_screen_refresh_rate(cookie);
+
+  PLUGINS_EVENT_HANDLE(event, randr_screen_change_notify, NULL);
+}
+
 /** Handler for KeyPress events reported once a key is pressed
  *
  * \param event The X KeyPress event
@@ -389,6 +422,7 @@ event_handle_configure_notify(xcb_configure_notify_event_t *event)
       globalconf.screen->width_in_pixels = event->width;
       globalconf.screen->height_in_pixels = event->height;
 
+      globalconf.background_reset = true;
       (*globalconf.rendering->reset_background)();
 
       return;
@@ -409,7 +443,10 @@ event_handle_configure_notify(xcb_configure_notify_event_t *event)
             a performance POV?
   */
   if(window_is_visible(window))
-    display_add_damaged_region(&window->region, true);
+    {
+      display_add_damaged_region(&window->region, true);
+      window->damaged_ratio = 1.0;
+    }
 
   /* Update geometry */
   window->geometry->x = event->x;
@@ -434,7 +471,7 @@ event_handle_configure_notify(xcb_configure_notify_event_t *event)
 
   if(window_is_visible(window))
     {
-      window->region = window_get_region(window);
+      window->region = window_get_region(window, true, false);
 
       if(update_pixmap)
         {
@@ -467,10 +504,10 @@ event_handle_create_notify(xcb_create_notify_event_t *event)
 
   /* Add  the  new window  whose  identifier  is  given in  the  event
      itself and  */
-  window_t *new_window = window_add(event->window);
+  window_t *new_window = window_add(event->window, false);
   if(!new_window)
     {
-      warn("Can't create window %jx", (uintmax_t) event->window);
+      debug("Cannot create window %jx", (uintmax_t) event->window);
       return;
     }
 
@@ -488,7 +525,7 @@ event_handle_create_notify(xcb_create_notify_event_t *event)
        creating regions all the time, this Region will be destroyed
        only upon DestroyNotify or re-created upon ConfigureNotify
     */
-    new_window->region = window_get_region(new_window);
+    new_window->region = window_get_region(new_window, true, true);
 
   PLUGINS_EVENT_HANDLE(event, create, new_window);
 }
@@ -542,7 +579,7 @@ event_handle_map_notify(xcb_map_notify_event_t *event)
 
   if(window_is_visible(window))
     {
-      window->region = window_get_region(window);
+      window->region = window_get_region(window, true, true);
 
       /* Everytime a window is mapped, a new pixmap is created */
       window_free_pixmap(window);
@@ -573,7 +610,7 @@ event_handle_reparent_notify(xcb_reparent_notify_event_t *event)
   /* Add the window if it is not already managed */ 
   if(event->parent == globalconf.screen->root ||
      !window_list_get(event->window))
-    window_add(event->window);
+    window_add(event->window, true);
   /* Don't manage the window if the parent is not the root window */
   else
     window_list_remove_window(window);
@@ -602,7 +639,10 @@ event_handle_unmap_notify(xcb_unmap_notify_event_t *event)
     }
 
   if(window_is_visible(window))
-    display_add_damaged_region(&window->region, true);
+    {
+      display_add_damaged_region(&window->region, true);
+      window->damaged_ratio = 1.0;
+    }
 
   /* Update window state */
   window->attributes->map_state = XCB_MAP_STATE_UNMAPPED;
@@ -629,6 +669,7 @@ event_handle_property_notify(xcb_property_notify_event_t *event)
      event->window == globalconf.screen->root)
     {
       debug("New background Pixmap set");
+      globalconf.background_reset = true;
       (*globalconf.rendering->reset_background)();
     }
 
@@ -691,11 +732,19 @@ event_handle(xcb_generic_event_t *event)
       event_handle_error((void *) event);
       return;
     }
-  else if(response_type == globalconf.extensions.damage->first_event + XCB_DAMAGE_NOTIFY)
+  else if(response_type == (globalconf.extensions.damage->first_event +
+                            XCB_DAMAGE_NOTIFY))
     {
       event_handle_damage_notify((void *) event);
       return;
     }
+  else if(globalconf.extensions.randr &&
+          response_type == (globalconf.extensions.randr->first_event +
+                            XCB_RANDR_SCREEN_CHANGE_NOTIFY))
+    {
+      event_handle_randr_screen_change_notify((void *) event);
+      return;
+    }
 
   switch(response_type)
     {
@@ -712,6 +761,7 @@ event_handle(xcb_generic_event_t *event)
       EVENT(XCB_UNMAP_NOTIFY, event_handle_unmap_notify);
       EVENT(XCB_PROPERTY_NOTIFY, event_handle_property_notify);
       EVENT(XCB_MAPPING_NOTIFY, event_handle_mapping_notify);
+#undef EVENT
     }
 }
 
