@@ -1,4 +1,5 @@
-/*
+/* -*-mode:c;coding:utf-8; c-basic-offset:2;fill-column:70;c-file-style:"gnu"-*-
+ *
  * Copyright (C) 2009 Arnaud "arnau" Fontaine <arnau@mini-dweeb.org>
  *
  * This  program is  free  software: you  can  redistribute it  and/or
@@ -24,6 +25,7 @@
 #include <stdint.h>
 #include <string.h>
 #include <assert.h>
+#include <math.h>
 
 #include <xcb/xcb.h>
 #include <xcb/xproto.h>
@@ -34,7 +36,7 @@
 #include "window.h"
 #include "structs.h"
 #include "atoms.h"
-#include "util.h"
+#include "display.h"
 
 /** Append a window to the end  of the windows list which is organized
  *  from the bottommost to the topmost window
@@ -63,34 +65,24 @@ window_list_append(const xcb_window_t new_window_id)
       window_tail->next = new_window;
     }
 
+  globalconf.windows_itree = util_itree_insert(globalconf.windows_itree,
+                                               new_window_id, new_window);
+
   return new_window;
 }
 
-/** Get the window object associated with the given Window XID
- *
- * \param window_id The Window XID to look for
- * \return The window object or NULL
- */
-window_t *
-window_list_get(const xcb_window_t window_id)
-{
-  window_t *window;
-
-  for(window = globalconf.windows;
-      window != NULL && window->id != window_id;
-      window = window->next)
-    ;
-
-  return window;
-}
-
 /** Free a given window and its associated resources
  *
  * \param window The window object to be freed
+ * \param do_itree_remove Should the window be removed from the itree as well
  */
 static void
-window_list_free_window(window_t *window)
+window_list_free_window(window_t *window, bool do_itree_remove)
 {
+  if(do_itree_remove)
+    globalconf.windows_itree = util_itree_remove(globalconf.windows_itree,
+                                                 window->id);
+
   /* Destroy the damage object if any */
   if(window->damage != XCB_NONE)
     {
@@ -128,14 +120,14 @@ window_list_remove_window(window_t *window_delete)
       window_t *old_window = globalconf.windows;
       globalconf.windows = globalconf.windows->next;
 
-      window_list_free_window(old_window);
+      window_list_free_window(old_window, true);
     }
   else
     for(window_t *window = globalconf.windows; window->next; window = window->next)
       if(window->next == window_delete)
 	{
 	  window->next = window->next->next;
-	  window_list_free_window(window_delete);
+	  window_list_free_window(window_delete, true);
 	  break;
 	}
 }
@@ -147,10 +139,17 @@ window_list_cleanup(void)
   window_t *window = globalconf.windows;
   window_t *window_next;
 
+  /* Destroy  the binary  tree,  values will  be  actually freed  when
+     clearing the linked list */
+  util_itree_free(globalconf.windows_itree);
+
   while(window != NULL)
     {
       window_next = window->next;
-      window_list_free_window(window);
+
+      /* Do not  remove it from the  itree as this is  already done by
+         util_itree_free() */
+      window_list_free_window(window, false);
       window = window_next;
     }
 }
@@ -288,6 +287,35 @@ window_get_pixmap(const window_t *window)
   return pixmap;
 }
 
+/** Check whether the given window is rectangular to optimize painting
+ *  as most windows are rectangular
+ *
+ * \param window The window object
+ * \return True if the window is rectangular
+ */
+bool
+window_is_rectangular(window_t *window)
+{
+  if(!window->shape_cookie.sequence)
+    return window->is_rectangular;
+
+  xcb_xfixes_fetch_region_reply_t *r =
+    xcb_xfixes_fetch_region_reply(globalconf.connection,
+                                  window->shape_cookie,
+                                  NULL);
+  if(r)
+    {
+      window->is_rectangular = xcb_xfixes_fetch_region_rectangles_length(r) <= 1;
+      free(r);
+    }
+  else
+    window->is_rectangular = true;
+
+  window->shape_cookie.sequence = 0;
+
+  return window->is_rectangular;
+}
+
 /** No need to include Shape extension header just for that */
 #define XCB_SHAPE_SK_BOUNDING 0
 
@@ -299,7 +327,7 @@ window_get_pixmap(const window_t *window)
  * \return The region associated with the given Window
  */
 xcb_xfixes_region_t
-window_get_region(window_t *window)
+window_get_region(window_t *window, bool screen_relative, bool check_shape)
 {
   xcb_xfixes_region_t new_region = xcb_generate_id(globalconf.connection);
 
@@ -308,15 +336,25 @@ window_get_region(window_t *window)
                                        window->id,
                                        XCB_SHAPE_SK_BOUNDING);
 
-  xcb_xfixes_translate_region(globalconf.connection,
-                              new_region,
-                              (int16_t) (window->geometry->x +
-                                         window->geometry->border_width),
-                              (int16_t) (window->geometry->y +
-                                         window->geometry->border_width));
+
+  if(screen_relative)
+    xcb_xfixes_translate_region(globalconf.connection,
+                                new_region,
+                                (int16_t) (window->geometry->x +
+                                           window->geometry->border_width),
+                                (int16_t) (window->geometry->y +
+                                           window->geometry->border_width));
 
   debug("Created new region %x from window %x", new_region, window->id);
 
+  if(check_shape)
+    {
+      window->shape_cookie = xcb_xfixes_fetch_region_unchecked(globalconf.connection,
+                                                               new_region);
+
+      xcb_flush(globalconf.connection);
+    }
+
   return new_region;
 }
 
@@ -388,39 +426,57 @@ window_get_invisible_window_pixmap_finalise(window_t *window)
   window_set_override_redirect(window, false);
 }
 
+typedef struct
+{
+  xcb_get_window_attributes_cookie_t attributes;
+  xcb_get_geometry_cookie_t geometry;
+} window_add_requests_cookies_t;
+
 /** Send requests when a window is added (CreateNotify or on startup),
- *  e.g. GetWindowAttributes request
+ *  e.g. GetWindowAttributes request and  GetGeometry if specified (in
+ *  most cases,  it is necessary  except on CreateNotify event  as the
+ *  geometry is already given in the event itself)
  *
  * \param window_id The Window XID
+ * \param get_geometry Should GetGeometry request be sent
  * \return The cookie associated with the request
  */
-static xcb_get_window_attributes_cookie_t
-window_add_requests(const xcb_window_t window_id)
+static window_add_requests_cookies_t
+window_add_requests(const xcb_window_t window_id, bool get_geometry)
 {
-  return xcb_get_window_attributes_unchecked(globalconf.connection,
-					     window_id);
+  window_add_requests_cookies_t cookies;
+  cookies.geometry.sequence = 0;
+
+  cookies.attributes = xcb_get_window_attributes(globalconf.connection,
+                                                 window_id);
+
+  cookies.geometry = xcb_get_geometry(globalconf.connection, window_id);
+
+  return cookies;
 }
 
-/** Get  the GetWindowAttributes  reply and  also associated  a Damage
- *  object  to it and  set the  attributes field  of the  given window
+/** Get  the GetWindowAttributes  and GetGeometry  (if requested  when
+ *  calling window_add_requests) replies and  also associated a Damage
+ *  object to  it and  set the  attributes field  of the  given window
  *  object
  *
+ * \see window_add_requests
  * \param window The window object
  * \param attributes_cookie The cookie associated with the GetWindowAttributes request
  * \return The GetWindowAttributes reply
  */
-static xcb_get_window_attributes_reply_t *
+static bool
 window_add_requests_finalise(window_t * const window,
-			     const xcb_get_window_attributes_cookie_t attributes_cookie)
+			     const window_add_requests_cookies_t window_add_cookies)
 {
   window->attributes = xcb_get_window_attributes_reply(globalconf.connection,
-						       attributes_cookie,
+						       window_add_cookies.attributes,
 						       NULL);
 
   if(!window->attributes)
     {
-      warn("GetWindowAttributes failed for window %jx", (uintmax_t) window->id);
-      return NULL;
+      debug("GetWindowAttributes failed for window %jx", (uintmax_t) window->id);
+      return false;
     }
 
   /* No  need to create  a Damage  object for  an InputOnly  window as
@@ -431,11 +487,29 @@ window_add_requests_finalise(window_t * const window,
     {
       window->damage = xcb_generate_id(globalconf.connection);
 
+      /* With DamageReportRawRectangles level, no attempt to compress
+         out overlapping rectangles is made, therefore many events are
+         received and handled needlessly, whereas with DamageReportNonEmpty
+         level only a single event specifying the full window region is sent
+         thus this is not efficient for small damage regions */
       xcb_damage_create(globalconf.connection, window->damage, window->id,
-			XCB_DAMAGE_REPORT_LEVEL_NON_EMPTY);
+			XCB_DAMAGE_REPORT_LEVEL_DELTA_RECTANGLES);
     }
 
-  return window->attributes;
+  if(window_add_cookies.geometry.sequence)
+    {
+      window->geometry = xcb_get_geometry_reply(globalconf.connection,
+                                                window_add_cookies.geometry,
+                                                NULL);
+
+      if(!window->geometry)
+        {
+          debug("GetGeometry failed for window %jx", (uintmax_t) window->id);
+          return false;
+        }
+    }
+
+  return true;
 }
 
 /** Manage  all  existing   windows  and  get  information  (geometry,
@@ -449,25 +523,15 @@ void
 window_manage_existing(const int nwindows,
 		       const xcb_window_t * const new_windows_id)
 {
-  xcb_get_window_attributes_cookie_t attributes_cookies[nwindows];
-  xcb_get_geometry_cookie_t geometry_cookies[nwindows];
+  window_add_requests_cookies_t window_add_cookies[nwindows];
 
   for(int nwindow = 0; nwindow < nwindows; ++nwindow)
-    {
-      /* Ignore the CM window */
-      if(new_windows_id[nwindow] == globalconf.cm_window)
-	continue;
+    /* Ignore the CM window */
+    if(new_windows_id[nwindow] != globalconf.cm_window)
+      window_add_cookies[nwindow] = window_add_requests(new_windows_id[nwindow],
+                                                        true);
 
-      attributes_cookies[nwindow] =
-	window_add_requests(new_windows_id[nwindow]);
-
-      /* Only  necessary  when adding  all  the  windows, otherwise  a
-	 window  is only  added on  CreateNotify event  which actually
-	 contains the window geometry */
-      geometry_cookies[nwindow] =
-	xcb_get_geometry_unchecked(globalconf.connection,
-				   new_windows_id[nwindow]);
-    }
+  globalconf.windows_itree = util_itree_new();
 
   window_t *new_windows[nwindows];
   for(int nwindow = 0; nwindow < nwindows; ++nwindow)
@@ -480,17 +544,13 @@ window_manage_existing(const int nwindows,
 	continue;
 
       if(!window_add_requests_finalise(new_windows[nwindow],
-					attributes_cookies[nwindow]))
+					window_add_cookies[nwindow]))
 	{
+          warn("Cannot manage window %jx", (uintmax_t) new_windows_id[nwindow]);
 	  window_list_remove_window(new_windows[nwindow]);
 	  continue;
 	}
 
-      new_windows[nwindow]->geometry =
-	xcb_get_geometry_reply(globalconf.connection,
-			       geometry_cookies[nwindow],
-			       NULL);
-
       /* The opacity  property is only  meaningful when the  window is
 	 mapped, because when the window is unmapped, we don't receive
 	 PropertyNotify */
@@ -503,7 +563,8 @@ window_manage_existing(const int nwindows,
           /* Get the Window Region as  well, this is also performed in
              CreateNotify   and   ConfigureNotify   handler  for   new
              Windows */
-          new_windows[nwindow]->region = window_get_region(new_windows[nwindow]);
+          new_windows[nwindow]->region = window_get_region(new_windows[nwindow],
+                                                           true, true);
 	}
     }
 
@@ -514,22 +575,23 @@ window_manage_existing(const int nwindows,
 
 /** Add  the  given   window  to  the  windows  list   and  also  send
  *  GetWindowAttributes request  (no need  to split this  function for
- *  now)
+ *  now) and GetGeometry if specified
  *
+ * \see window_add_requests
  * \param new_window_id The new Window XID
+ * \param get_geometry Should GetGeometry request be sent
  * \return The new window object
  */
 window_t *
-window_add(const xcb_window_t new_window_id)
+window_add(const xcb_window_t new_window_id, bool get_geometry)
 {
-  xcb_get_window_attributes_cookie_t attributes_cookie =
-    window_add_requests(new_window_id);
+  window_add_requests_cookies_t cookies = window_add_requests(new_window_id,
+                                                              get_geometry);
 
   window_t *new_window = window_list_append(new_window_id);
 
   /* The request should never fail... */
-  if(!window_add_requests_finalise(new_window,
-				    attributes_cookie))
+  if(!window_add_requests_finalise(new_window, cookies))
     {
       window_list_remove_window(new_window);
       return NULL;
@@ -595,19 +657,41 @@ window_restack(window_t *window, xcb_window_t window_new_above_id)
 void
 window_paint_all(window_t *windows)
 {
+  /* If the background  is reset, then repaint the  whole screen, it's
+     bad from a performance point of view, but it's done rarely */
+  if(globalconf.background_reset)
+    display_reset_damaged();
+
   (*globalconf.rendering->paint_background)();
 
   for(window_t *window = windows; window; window = window->next)
     {
-      if(!window->damaged || !window_is_visible(window))
-	{
-	  debug("Ignoring window %jx", (uintmax_t) window->id);
-	  continue;
-	}
+      if(window->damaged)
+        {
+          debug("Painting window %jx", (uintmax_t) window->id);
+          (*globalconf.rendering->paint_window)(window);
+        }
+      /* When the  window has been damaged  or was damaged but  is not
+         visible anymore */
+      if(window->damaged_ratio)
+        {
+          /* Reset damaged ratio for the next repaint */
+          window->damaged_ratio = 0.0;
 
-      debug("Painting window %jx", (uintmax_t) window->id);
-      (*globalconf.rendering->paint_window)(window);
+          /* And the DamageNotify events counter */
+          window->damage_notify_counter = 0;
+
+          /* Reset  the  damaged  region   in  order  to  get  damages
+             occurring    after   the    repaint,   otherwise,    with
+             DamageReportDeltaRectangles level,  DamageNotify won't be
+             send if  the same region  was already damaged  during the
+             previous repaint */
+          xcb_damage_subtract(globalconf.connection, window->damage,
+                              XCB_NONE, XCB_NONE);
+        }
     }
 
   (*globalconf.rendering->paint_all)();
+  globalconf.background_reset = false;
+  xcb_aux_sync(globalconf.connection);
 }
